/* 	Copyright John D. "smeezekitty" 2017
	
	FreeNBTV is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FreeNBTV is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FreeNBTV.  If not, see <http://www.gnu.org/licenses/>.
*/

/* This file handles the GUI for the output setup*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <gtk/gtk.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include "nbtv_format.h"
#include "gui.h"
#include "input.h"
#include "nbtv_gen.h"
#include "output.h"

int o_handler_disable; //Disable the change callbacks

GtkWidget *output_window;
GtkWidget *output_grid;
GtkWidget *output_to_file;
GtkWidget *output_to_device;
GtkWidget *main_frame;
GtkWidget *mf_grid;
GtkWidget *main_fn;
GtkWidget *wave_fmt;
GtkWidget *raw_fmt;
GtkWidget *audio_device;
GtkWidget *channel_count;
GtkWidget *LPF;
GtkWidget *time_box;
GtkWidget *run_seconds;
GtkWidget *bit_depth;
GtkWidget *sample_rate;
GtkWidget *start_stop;
GtkWidget *time_produced,*tp_bar;
GtkWidget *output_tabs;
GtkWidget *tab_frame;
GtkWidget *signed_s;
GtkWidget *sf_box;
GtkWidget *g_box;
GtkWidget *set_format;
/*
struct output_tab {
	GtkWidget *tgrid;
	GtkWidget *null_output;
	GtkWidget *file_output;
	GtkWidget *main_output;
	GtkWidget *main_channel;
	GtkWidget *file_name;
	GtkWidget *raw_file;
	GtkWidget *wav_file;
} otabs[4];
*/
int sample_rates[]={8000, 11025, 16000, 19200, 22050, 24000, 32000, 44100, 48000, 96000, 192000, -1};

void start_n(GtkWidget *w, gpointer bla){
	if(nbtv_gen){
		stop_NBTV_gen();
	} else {
		start_NBTV_gen(gtk_spin_button_get_value(GTK_SPIN_BUTTON(run_seconds)), &input, nbtv_fmt);
	}
	gui_update_output(&output);
}
GtkWidget *create_sample_rate_combobox(){
	GtkWidget *cb = gtk_combo_box_text_new();
	char name[16];
	int i;
	for(i = 0;i < sample_rates[i]>0;i++){
		itoa(sample_rates[i], name, 10);
		gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(cb), NULL, name);
	}
	gtk_combo_box_set_active(GTK_COMBO_BOX(cb), 7);
	return cb;
}
GtkWidget *create_bps_combobox(){
	GtkWidget *cb = gtk_combo_box_text_new();
	char name[16];
	int i;
	for(i = 4;i < 64;i<<=1){
		itoa(i, name, 10);
		gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(cb), NULL, name);
	}
	gtk_combo_box_set_active(GTK_COMBO_BOX(cb), 1);
	return cb;
}
GtkWidget *create_device_combobox(){
	/*We don't enumerate audio devices yet*/
	GtkWidget *cb = gtk_combo_box_text_new();
	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(cb), NULL, "<DEFAULT>");
	gtk_combo_box_set_active(GTK_COMBO_BOX(cb), 0);
	return cb;
}
/*void gui_output_create_tabs(){
	struct output_tab *t;
	char buffer[64];
	output_tabs = gtk_notebook_new();
	tab_frame = gtk_frame_new("Channels: ");
	gtk_container_add(GTK_CONTAINER(tab_frame), output_tabs);
	int i;
	for(i = 0;i < 4;i++){
		t = &otabs[i];
		sprintf(buffer, "Channel %d", i);
		t->tgrid = gtk_grid_new();
		gtk_notebook_append_page(GTK_NOTEBOOK(output_tabs), t->tgrid, gtk_label_new(buffer));
		t->null_output = gtk_radio_button_new_with_label(NULL, "Null Output");
		t->main_output = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(t->null_output), "Main Output\nChannel: ");
		t->file_output = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(t->null_output), "File Output: ");
		t->main_channel = gtk_combo_box_text_new();
		t->file_name = gtk_entry_new();
		gtk_grid_attach(GTK_GRID(t->tgrid), t->null_output, 0, 0, 1, 1);
		gtk_grid_attach(GTK_GRID(t->tgrid), t->main_output, 0, 1, 1, 1);
		gtk_grid_attach(GTK_GRID(t->tgrid), t->file_output, 0, 2, 1, 1);
		gtk_grid_attach(GTK_GRID(t->tgrid), t->main_channel, 1, 1, 1, 1);
		gtk_grid_attach(GTK_GRID(t->tgrid), t->file_name, 1, 2, 1, 1);
	}
	gtk_grid_attach(GTK_GRID(output_grid), tab_frame, 1, 0, 1, 1);
}*/
void sof(GtkWidget *w, gpointer bla){
	set_output_format(&output);
}
void gui_output_connect_signals(){
	g_signal_connect(output_window, "destroy", G_CALLBACK(destroy), NULL);
	g_signal_connect(start_stop, "clicked", G_CALLBACK(start_n), NULL);
	g_signal_connect(set_format, "clicked", G_CALLBACK(sof), NULL);
}
void gui_create_output_window(){
	output_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_resizable(GTK_WINDOW(output_window), FALSE);
	output_grid = gtk_grid_new();
	gtk_container_add(GTK_CONTAINER(output_window), output_grid);
	main_frame = gtk_frame_new("Primary output: ");
	mf_grid = gtk_grid_new();
	gtk_grid_attach(GTK_GRID(output_grid), main_frame, 0, 0, 1, 1);
	gtk_container_add(GTK_CONTAINER(main_frame), mf_grid);
	output_to_device = gtk_radio_button_new_with_label(NULL, "Audio Device:");
	gtk_grid_attach(GTK_GRID(mf_grid), output_to_device, 0, 0, 1, 1);
	output_to_file = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(output_to_device), "File:");
	gtk_grid_attach(GTK_GRID(mf_grid), output_to_file, 0, 1, 1, 1);
	main_fn = gtk_entry_new();
	gtk_grid_attach(GTK_GRID(mf_grid), main_fn, 1, 1, 1, 1);
	wave_fmt = gtk_radio_button_new_with_label(NULL, "WAVE format");
	raw_fmt = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(wave_fmt), "RAW format");
	gtk_grid_attach(GTK_GRID(mf_grid), wave_fmt, 2, 1, 1, 1);
	gtk_grid_attach(GTK_GRID(mf_grid), raw_fmt, 3, 1, 1, 1);
	time_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 4);
	put_label(mf_grid, "Sample Rate:", 0, 2);
	//put_label(mf_grid, "Bits Per Sample:", 2, 2); 
	put_label(mf_grid, "Channel Count:", 0, 3);
	//put_label(mf_grid, "LPF Amount", 2, 3);
	put_label(mf_grid, "Produce ", 0, 4);
	sample_rate = create_sample_rate_combobox();
	bit_depth = create_bps_combobox();
	sf_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 4);
	g_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 4);
	gtk_grid_attach(GTK_GRID(mf_grid), sf_box, 1, 2, 1, 1);
	gtk_grid_attach(GTK_GRID(mf_grid), g_box, 1, 3, 1, 1);
//	gtk_grid_attach(GTK_GRID(mf_grid), sample_rate, 1, 2, 1, 1);
//	gtk_grid_attach(GTK_GRID(mf_grid), bit_depth, 3, 2, 1, 1);
	run_seconds = gtk_spin_button_new_with_range(0, 3600, 1);
	gtk_box_pack_start(GTK_BOX(time_box), run_seconds, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(time_box), gtk_label_new("Seconds\n(0 for unbounded)"), TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(sf_box), sample_rate, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(sf_box), gtk_label_new("BPS:"), TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(sf_box), bit_depth, TRUE, TRUE, 0);
	signed_s = gtk_check_button_new_with_label("Signed");
	gtk_grid_attach(GTK_GRID(mf_grid), signed_s, 2, 2, 1, 1);
	gtk_grid_attach(GTK_GRID(mf_grid), time_box, 1, 4, 1, 1);
	time_produced = gtk_label_new("0:00:00.00");
	gtk_grid_attach(GTK_GRID(mf_grid), time_produced, 2, 4, 1, 1);
	start_stop = gtk_button_new_with_label("START");
	gtk_grid_attach(GTK_GRID(mf_grid), start_stop, 3, 4, 1, 1);
	tp_bar = gtk_progress_bar_new();
	gtk_grid_attach(GTK_GRID(mf_grid), tp_bar, 0, 5, 5, 1);
	audio_device = create_device_combobox();
	gtk_grid_attach(GTK_GRID(mf_grid), audio_device, 1, 0, 1, 1);
	channel_count = gtk_spin_button_new_with_range(1, 4, 1);
	gtk_box_pack_start(GTK_BOX(g_box), channel_count, FALSE, TRUE, 0);
	//gtk_grid_attach(GTK_GRID(mf_grid), channel_count, 1, 3, 1, 1);
	LPF = gtk_scale_new_with_range(GTK_ORIENTATION_HORIZONTAL, 0, 5, 1);
	gtk_scale_set_draw_value(GTK_SCALE(LPF), FALSE);
//	gtk_grid_attach(GTK_GRID(mf_grid), LPF, 3, 3, 1, 1);
	gtk_box_pack_start(GTK_BOX(g_box), gtk_label_new("LPF: "), FALSE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(g_box), LPF, TRUE, TRUE, 0);
	set_format = gtk_button_new_with_label("Set Format");
	gtk_grid_attach(GTK_GRID(mf_grid), set_format, 2, 3, 2, 1);
//	gui_output_create_tabs();
	gui_output_connect_signals();
	gtk_widget_show_all(output_window);
}
/*Converts the timestamp in milliseconds to hours, minutes, seconds and hundredths of seconds*/
void gui_update_time_produced(){
	char buffer[16];
	int hr,min,sec,hs;
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(tp_bar), (double)output_timestamp/(double)target_timestamp);
	int timestamp = output_timestamp;
	hr = timestamp / 3600000;
	timestamp -= hr * 3600000;
	min = timestamp / 60000;
	timestamp -= min * 60000;
	sec = timestamp  / 1000;
	timestamp -= sec * 1000;
	hs = timestamp / 10;
	sprintf(buffer, "%d:%02d:%02d.%02d", hr, min, sec, hs);
	gtk_label_set_text(GTK_LABEL(time_produced), buffer);
}
void gui_update_output(struct _output *op){
	if(!output_window) return;
	o_handler_disable = 1;
	gtk_combo_box_set_active(GTK_COMBO_BOX(audio_device), op->device);
	gtk_entry_set_text(GTK_ENTRY(main_fn), op->main_filename);
	int sr;
	switch(op->sample_rate){
		default:
		case 8000:
			sr = 0;
			break;
		case 11025:
			sr = 1;
			break;
		case 16000:
			sr = 2;
			break;
		case 19200:
			sr = 3;
			break;
		case 22050:
			sr = 4;
			break;
		case 24000:
			sr = 5;
			break;
		case 32000:
			sr = 6;
			break;
		case 44100:
			sr = 7;
			break;
		case 48000:
			sr = 8;
			break;
		case 96000:
			sr = 9;
			break;
		case 192000:
			sr = 10;
			break;
	}
	gtk_combo_box_set_active(GTK_COMBO_BOX(sample_rate), sr);
	gtk_combo_box_set_active(GTK_COMBO_BOX(bit_depth), op->bps/9.0+0.25);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(signed_s), op->signed_samples);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(channel_count), op->num_channels);
	gtk_range_set_value(GTK_RANGE(LPF), op->lpf);
	//gtk_spin_button_set_value(GTK_SPIN_BUTTON(run_seconds), op->target_frame_count/nbtv_fmt->frame_rate);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(wave_fmt), op->format == OUTPUT_FORMAT_WAVE);
	//gtk_label_set_text(GTK_LABEL(time_produced), op->current_frame/nbtv_fmt->frame_rate);
	if(op->format == OUTPUT_FORMAT_SC){
		gtk_widget_set_sensitive(audio_device, 1);
		gtk_widget_set_sensitive(main_fn, 0);
		gtk_widget_set_sensitive(wave_fmt, 0);
		gtk_widget_set_sensitive(raw_fmt, 0);
	} else {
		gtk_widget_set_sensitive(audio_device, 0);
		gtk_widget_set_sensitive(main_fn, 1);
		gtk_widget_set_sensitive(wave_fmt, 1);
		gtk_widget_set_sensitive(raw_fmt, 1);
	}
	gui_update_time_produced(op);
	/*int i;
	for(i = 0;i < 4;i++){
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(otabs[i].null_output), op->channels[i].format == 0);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(otabs[i].main_output), op->channels[i].format == OUTPUT_FORMAT_MAIN);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(otabs[i].file_output), op->channels[i].format == OUTPUT_FORMAT_RAW || op->channels[i].format == OUTPUT_FORMAT_WAVE);
		gtk_entry_set_text(GTK_ENTRY(otabs[i].file_name), op->channels[i].filename);
	}*/
	if(nbtv_gen){
		gtk_button_set_label(GTK_BUTTON(start_stop), "STOP");
	} else {
		gtk_button_set_label(GTK_BUTTON(start_stop), "START");
	}
	o_handler_disable = 0;
}
void set_output_format(struct _output *op){
	char *ext = 0;
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(output_to_file))){
		strcpy(op->main_filename, gtk_entry_get_text(GTK_ENTRY(main_fn)));
		/*Make sure we have a filename if we're outputing to a file. Normally we base this on the input filename if available*/
		if(!*op->main_filename || !strcmp(op->main_filename, ".wav")){
			if(!*input.filename){
				strcpy(op->main_filename, "output");
			} else {
				strcpy(op->main_filename, basename(input.filename));
				ext = strrchr(op->main_filename, '.');
				*ext = 0;
			}
		}
		if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wave_fmt))){
			op->format = OUTPUT_FORMAT_WAVE;
			if(ext) strcat(op->main_filename, ".wav");
		} else {
			op->format = OUTPUT_FORMAT_RAW;
		}
	} else {
		op->format = OUTPUT_FORMAT_SC;
	}
	op->sample_rate = atoi(gtk_combo_box_text_get_active_text(GTK_COMBO_BOX_TEXT(sample_rate)));
	op->bps = atoi(gtk_combo_box_text_get_active_text(GTK_COMBO_BOX_TEXT(bit_depth)));
	op->signed_samples = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(signed_s));
	op->num_channels = gtk_spin_button_get_value(GTK_SPIN_BUTTON(channel_count));
	op->lpf = gtk_range_get_value(GTK_RANGE(LPF));
	op->device = gtk_combo_box_get_active(GTK_COMBO_BOX(audio_device));
	if(op->bps > 8 && op->format != OUTPUT_FORMAT_RAW){
		op->signed_samples = 1; //Portaudio and MS-WAVE files only support signed samples with >8bps
	}
	setup_output(op);
	int i = 0;
	int p = 0;
	/*for(i = 0;i < 4;i++){
		op->channels[i].format = 0;
		if(!(nbtv_fmt->channels[i].sync || nbtv_fmt->channels[i].video || nbtv_fmt->channels[i].audio) || p > op->num_channels){
			continue;
		}
		op->channels[i].main_channel_num = p++;
		op->channels[i].format = OUTPUT_FORMAT_MAIN;
		
	}*/
	gui_update_output(op);
}
	
